//********************************************************************
//*** Thread/TaskRunner.cc
//*** Copyright (c) 2003-2008 by Markus Winand <mws@fatalmind.com>
//*** $Id: TaskRunner.cc,v 1.16 2008/01/27 16:42:29 mws Exp $
//********************************************************************
/*
This file is part of ResourcePool.

ResourcePool is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2 of 
the License, or (at your option) any later version.

ResourcePool is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ResourcePool; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307  USA */

#ifndef INCLUDED_TASKRUNNER_HH
#include "TaskRunner.hh"
#endif

#include "Thread/WaitObject.hh"

using namespace std;

namespace fatalmind {

/*
 *
 * Declaration of TaskRunnerTask
 *
 */

class TaskRunnerTask: public Task {
	public:
		TaskRunnerTask() 
		: _tsk()
		, _condvar()
		, _stopped(false)
		, _wo()
		{
		};
		virtual ~TaskRunnerTask() {
		};

		virtual void run();
	
		void stop();

	protected:
		void waitForTsk();
		void waitForSlot();
		void runDetachedLocked(CommonPtr<Task, MultiThreadedModel<> >&);

		void runDetached(CommonPtr<Task, MultiThreadedModel<> >& tsk);
		TaskRunnerInterface::ID run(CommonPtr<Task, MultiThreadedModel<> >& tsk);

	friend 
		class TaskRunner;

	private:
		typedef FastLock<>  lock_t;
		typedef Synchronize<lock_t> sync_t;

		CommonPtr<Task, MultiThreadedModel<> >   _tsk;

		CondVar<lock_t>   _condvar;	
		bool              _stopped;
		gc_ptr<WaitObject, MultiThreadedModel<> > _wo;
};

/*
 *
 * Implementation of TaskRunner
 *
 */

TaskRunner::Stopped::Stopped(string msg)
: Exception(msg)
{
}

TaskRunner::TaskRunner()
: _tsk(new TaskRunnerTask())
{
	start(_tsk.get());
}

TaskRunner::~TaskRunner()
{
	try {
		stop();
	} catch (...) {
	}
}


void TaskRunner::runDetached(CommonPtr<Task, MultiThreadedModel<> > tsk) {
	_tsk->runDetached(tsk);
}

TaskRunnerInterface::ID TaskRunner::run(CommonPtr<Task, MultiThreadedModel<> > tsk) {
	return _tsk->run(tsk);
}

void TaskRunner::stop() {
	_tsk->stop();
	join();
}

/*
 *
 * TaskRunnerTask
 *
 */

void TaskRunnerTask::run() {
	sync_t lk(_condvar);
	waitForTsk();
	while (! _stopped) {
		try {
			_tsk->run();
		} catch (...) {
			// TODO:
		}
		if (_wo.get() != 0) {
			_wo->done();
		}
		_wo.reset();
		_tsk.reset();
		_condvar.signal();
		waitForTsk();
	};
}

void TaskRunnerTask::waitForTsk() {	
	while (_tsk.get() == 0 && !_stopped) {
		_condvar.wait();
	}
}

void TaskRunnerTask::waitForSlot() {	
	while (_tsk.get() != 0 && !_stopped) {
		_condvar.wait();
	}
	if (_stopped) {
		throw TaskRunner::Stopped("The TaskRunner has already been stopped");
	}
}

void TaskRunnerTask::stop() {
	try {
		sync_t lk(_condvar);
		waitForSlot();
		_stopped = true;
		_condvar.broadcast();
	} catch (...) {
	}
}

void TaskRunnerTask::runDetached(CommonPtr<Task, MultiThreadedModel<> >& tsk) 
{
	sync_t lk(_condvar);
	waitForSlot();
	_wo.reset();
	runDetachedLocked(tsk);
}

void TaskRunnerTask::runDetachedLocked(CommonPtr<Task, MultiThreadedModel<> >& tsk) 
{
	_tsk = tsk;
	_condvar.broadcast();
}

TaskRunnerInterface::ID TaskRunnerTask::run(CommonPtr<Task, MultiThreadedModel<> >& tsk) {
	sync_t lk(_condvar);
	waitForSlot();
	gc_ptr<WaitObject, MultiThreadedModel<> > wc(new WaitObject());
	_wo = wc;
	TaskRunnerInterface::ID id(_wo);
	runDetachedLocked(tsk);
	return id;
}

} // namespace fatalmind
